%
% MTL parser for ETM+ dataset
%
% Written in Mar 2011 by Seongsu Jeong
% Geomatics and Remote Sensing Laboratory (GRSLAB), Yonsei University
%
% MAIN FUNCTIONALITY:
% Reads the information from MTL text file, which is provided with LANDSAT 7 ETM+ data.
%
% USAGE:
% MTL=MTL_parser(filename) - for a single MTL file
% MTL_list=MTL_parser(filename_list) - for multiple MTL file
%                                      The list should be in a matrix form
%                                      which each of its row vector is a
%                                      string of a MTL filename
% MTL_list=MTL_parser() - Searches all MTL files in current directory and
%                         parses them all
% 
% NOTIFICATION FOR THE OUTPUT VARIABLE:
% The values which surrounded "double quotation mark" was treated as a
% string, along with the information related to the time and date. Other
% variables were cosidered as a double-precision floating point variables.
%
% UPDATED FEATURES:
% 08-MAR-2011: Capable of dealing with void input
%              When this case occurs, this module searches all MTL files
%              based on its filename and list them up. After then, this
%              function reads all detedted MTL files and returns the result
%              as a structure array in column vector form.
%
% LAST MODIFICATION: 08-MAR-2011
% 



function MTL_list=MTL_parser(MTL_filename)

nMTL=0;
if nargin()==0 %when there is no input argument    
    file_list=ls;
    for cnt=1:size(file_list,1) %search all MTL files based on the file name
        if ~isempty(findstr(file_list(cnt,:),'MTL.txt'))
            nMTL=nMTL+1;
            MTL_filename(nMTL,:)=file_list(cnt,:);
        end
    end
    
elseif nargin()==1    
    nMTL=size(MTL_filename,1);    
    
else
    error('Incorrect input argument: Input a name of (or a list of) filename or make the input argument to void');
    
end

%% Begin the processing for all available MTL files
for cnt=1:nMTL
    % File open and line string input
    fin=fopen(MTL_filename(cnt,:),'r');
    str_in=fgetl(fin);

    while ~strcmp(str_in,'END')
        % String parsing and refinement
        % input line refinement and processing
        
        [field value]=strtok(str_in,'=');
        
        %field name refinement
        %remove unnecessary space character from the field description
        if sum(field(1)==' ') %space character detector
                field=strtok(field,' ');
                
        end
        
        %Value refinement
        value=strtok(value,'=');
        
        %remove unnecessary space character from the field description
        while value(1,1)==' '
            value=value(1,2:size(value,2));
            
        end

        %disp(strcat(field,': ',value));
        
        if sum(value(1)=='"' && value(size(value,2))=='"') %If the value is a string wrapped by large quotation mark (")
            value=value(1,2:size(value,2)-1);
            
        elseif isempty(findstr(field,'TIME')) && isempty(findstr(field,'DATE'))
            value=str2num(value);
        
        end
        
        % Field detection and assignment routine
        switch field
            
            % GROUP=METADATA_FILE_INFO
            case 'ORIGIN'
                MTL.METADATA_FILE_INFO.ORIGIN=value;                
            case 'REQUEST_ID'
                MTL.METADATA_FILE_INFO.REQUEST_ID=value;
            case 'PRODUCT_CREATION_TIME'
                MTL.METADATA_FILE_INFO.PRODUCT_CREATION_TIME=value;
            case 'STATION_ID'
                MTL.METADATA_FILE_INFO.STATION_ID=value;
            case 'LANDSAT7_XBAND'
                MTL.METADATA_FILE_INFO.LANDSAT7_XBAND=value;
            case 'GROUND_STATION'
                MTL.METADATA_FILE_INFO.GROUND_STATION=value;
            case 'LPS_PROCESSOR_NUMBER'
                MTL.METADATA_FILE_INFO.LPS_PROCESSOR_NUMBER=value;
            case 'DATEHOUR_CONTACT_PERIOD'
                MTL.METADATA_FILE_INFO.DATEHOUR_CONTACT_PERIOD=value;
            case 'SUBINTERVAL_NUMBER'
                MTL.METADATA_FILE_INFO.SUBINTERVAL_NUMBER=value;
 
            % GROUP=PRODUCT_METADATA
            case 'PRODUCT_TYPE'        
                MTL.PRODUCT_METADATA.PRODUCT_TYPE=value;
            case 'ELEVATION_SOURCE'        
                MTL.PRODUCT_METADATA.ELEVATION_SOURCE=value;
            case 'PROCESSING_SOFTWARE'        
                MTL.PRODUCT_METADATA.PROCESSING_SOFTWARE=value;
            case 'EPHEMERIS_TYPE'        
                MTL.PRODUCT_METADATA.EPHEMERIS_TYPE=value;
            case 'SPACECRAFT_ID'        
                MTL.PRODUCT_METADATA.SPACECRAFT_ID=value;
            case 'SENSOR_ID'        
                MTL.PRODUCT_METADATA.SENSOR_ID=value;
            case 'SENSOR_MODE'        
                MTL.PRODUCT_METADATA.SENSOR_MODE=value;
            case 'ACQUISITION_DATE'        
                MTL.PRODUCT_METADATA.ACQUISITION_DATE=value;
            case 'SCENE_CENTER_SCAN_TIME'        
                MTL.PRODUCT_METADATA.SCENE_CENTER_SCAN_TIME=value;
            case 'WRS_PATH'        
                MTL.PRODUCT_METADATA.WRS_PATH=value;
            case 'STARTING_ROW'        
                MTL.PRODUCT_METADATA.STARTING_ROW=value;
            case 'ENDING_ROW'        
                MTL.PRODUCT_METADATA.ENDING_ROW=value;
            case 'BAND_COMBINATION'        
                MTL.PRODUCT_METADATA.BAND_COMBINATION=value;
            case 'PRODUCT_UL_CORNER_LAT'        
                MTL.PRODUCT_METADATA.PRODUCT_UL_CORNER_LAT=value;
            case 'PRODUCT_UL_CORNER_LON'        
                MTL.PRODUCT_METADATA.PRODUCT_UL_CORNER_LON=value;
            case 'PRODUCT_UR_CORNER_LAT'        
                MTL.PRODUCT_METADATA.PRODUCT_UR_CORNER_LAT=value;
            case 'PRODUCT_UR_CORNER_LON'        
                MTL.PRODUCT_METADATA.PRODUCT_UR_CORNER_LON=value;
            case 'PRODUCT_LL_CORNER_LAT'        
                MTL.PRODUCT_METADATA.PRODUCT_LL_CORNER_LAT=value;        
            case 'PRODUCT_LL_CORNER_LON'        
                MTL.PRODUCT_METADATA.PRODUCT_LL_CORNER_LON=value;
            case 'PRODUCT_LR_CORNER_LAT'        
                MTL.PRODUCT_METADATA.PRODUCT_LR_CORNER_LAT=value;
            case 'PRODUCT_LR_CORNER_LON'        
                MTL.PRODUCT_METADATA.PRODUCT_LR_CORNER_LON=value;
            case 'PRODUCT_UL_CORNER_MAPX'        
                MTL.PRODUCT_METADATA.PRODUCT_UL_CORNER_MAPX=value;
            case 'PRODUCT_UL_CORNER_MAPY'        
                MTL.PRODUCT_METADATA.PRODUCT_UL_CORNER_MAPY=value;
            case 'PRODUCT_UR_CORNER_MAPX'        
                MTL.PRODUCT_METADATA.PRODUCT_UR_CORNER_MAPX=value;
            case 'PRODUCT_UR_CORNER_MAPY'        
                MTL.PRODUCT_METADATA.PRODUCT_UR_CORNER_MAPY=value;
            case 'PRODUCT_LL_CORNER_MAPX'        
                MTL.PRODUCT_METADATA.PRODUCT_LL_CORNER_MAPX=value;
            case 'PRODUCT_LL_CORNER_MAPY'        
                MTL.PRODUCT_METADATA.PRODUCT_LL_CORNER_MAPY=value;
            case 'PRODUCT_LR_CORNER_MAPX'        
                MTL.PRODUCT_METADATA.PRODUCT_LR_CORNER_MAPX=value;
            case 'PRODUCT_LR_CORNER_MAPY'        
                MTL.PRODUCT_METADATA.PRODUCT_LR_CORNER_MAPY=value;     
            case 'PRODUCT_SAMPLES_PAN'        
                MTL.PRODUCT_METADATA.PRODUCT_SAMPLES_PAN=value;     
            case 'PRODUCT_LINES_PAN'        
                MTL.PRODUCT_METADATA.PRODUCT_LINES_PAN=value;     
            case 'PRODUCT_SAMPLES_REF'        
                MTL.PRODUCT_METADATA.PRODUCT_SAMPLES_REF=value;     
            case 'PRODUCT_LINES_REF'        
                MTL.PRODUCT_METADATA.PRODUCT_LINES_REF=value;     
            case 'PRODUCT_SAMPLES_THM'        
                MTL.PRODUCT_METADATA.PRODUCT_SAMPLES_THM=value;     
            case 'PRODUCT_LINES_THM'        
                MTL.PRODUCT_METADATA.PRODUCT_LINES_THM=value;     
            case 'BAND1_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND1_FILE_NAME=value;     
            case 'BAND2_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND2_FILE_NAME=value;     
            case 'BAND3_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND3_FILE_NAME=value;     
            case 'BAND4_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND4_FILE_NAME=value;     
            case 'BAND5_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND5_FILE_NAME=value;     
            case 'BAND61_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND61_FILE_NAME=value;
            case 'BAND62_FILE_NAME'        
                MTL.PRODUCT_METADATA.BAND62_FILE_NAME=value;
            case 'BAND7_FILE_NAME'
                MTL.PRODUCT_METADATA.BAND7_FILE_NAME=value;
            case 'BAND8_FILE_NAME'
                MTL.PRODUCT_METADATA.BAND8_FILE_NAME=value;
            case 'GCP_FILE_NAME'
                MTL.PRODUCT_METADATA.GCP_FILE_NAME=value;
            case 'METADATA_L1_FILE_NAME'
                MTL.PRODUCT_METADATA.METADATA_L1_FILE_NAME=value;
            case 'CPF_FILE_NAME'
                MTL.PRODUCT_METADATA.CPF_FILE_NAME=value;
   
            %GROUP = MIN_MAX_RADIANCE
            case 'LMAX_BAND1'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND1=value;        
            case 'LMIN_BAND1'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND1=value;        
            case 'LMAX_BAND2'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND2=value;        
            case 'LMIN_BAND2'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND2=value;        
            case 'LMAX_BAND3'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND3=value;        
            case 'LMIN_BAND3'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND3=value;        
            case 'LMAX_BAND4'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND4=value;        
            case 'LMIN_BAND4'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND4=value;        
            case 'LMAX_BAND5'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND5=value;        
            case 'LMIN_BAND5'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND5=value;        
            case 'LMAX_BAND61'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND61=value;        
            case 'LMIN_BAND61'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND61=value;        
            case 'LMAX_BAND62'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND62=value;        
            case 'LMIN_BAND62'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND62=value;        
            case 'LMAX_BAND7'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND7=value;        
            case 'LMIN_BAND7'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND7=value;        
            case 'LMAX_BAND8'        
                MTL.MIN_MAX_RADIANCE.LMAX_BAND8=value;        
            case 'LMIN_BAND8'        
                MTL.MIN_MAX_RADIANCE.LMIN_BAND8=value;        
    
            %GROUP = MIN_MAX_PIXEL_VALUE        
            case 'QCALMAX_BAND1'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND1=value;
            case 'QCALMIN_BAND1'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND1=value;
            case 'QCALMAX_BAND2'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND2=value;
            case 'QCALMIN_BAND2'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND2=value;
            case 'QCALMAX_BAND3'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND3=value;
            case 'QCALMIN_BAND3'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND3=value;
            case 'QCALMAX_BAND4'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND4=value;
            case 'QCALMIN_BAND4'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND4=value;
            case 'QCALMAX_BAND5'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND5=value;
            case 'QCALMIN_BAND5'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND5=value;
            case 'QCALMAX_BAND61'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND61=value;
            case 'QCALMIN_BAND61'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND61=value;
            case 'QCALMAX_BAND62'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND62=value;
            case 'QCALMIN_BAND62'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND62=value;
            case 'QCALMAX_BAND7'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND7=value;
            case 'QCALMIN_BAND7'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND7=value;
            case 'QCALMAX_BAND8'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMAX_BAND8=value;
            case 'QCALMIN_BAND8'
                MTL.MIN_MAX_PIXEL_VALUE.QCALMIN_BAND8=value;
        
            %GROUP = PRODUCT_PARAMETERS
            case 'CORRECTION_METHOD_GAIN_BAND1'
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND1=value;
            case 'CORRECTION_METHOD_GAIN_BAND2'
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND2=value;
            case 'CORRECTION_METHOD_GAIN_BAND3'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND3=value;
            case 'CORRECTION_METHOD_GAIN_BAND4'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND4=value;
            case 'CORRECTION_METHOD_GAIN_BAND5'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND5=value;
            case 'CORRECTION_METHOD_GAIN_BAND61'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND61=value;
            case 'CORRECTION_METHOD_GAIN_BAND62'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND62=value;
            case 'CORRECTION_METHOD_GAIN_BAND7'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND7=value;
            case 'CORRECTION_METHOD_GAIN_BAND8'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_GAIN_BAND8=value;
            case 'CORRECTION_METHOD_BIAS'    
                MTL.PRODUCT_PARAMETERS.CORRECTION_METHOD_BIAS=value;
            case 'BAND1_GAIN'    
                MTL.PRODUCT_PARAMETERS.BAND1_GAIN=value;        
            case 'BAND2_GAIN'    
                MTL.PRODUCT_PARAMETERS.BAND2_GAIN=value;        
            case 'BAND3_GAIN'    
                MTL.PRODUCT_PARAMETERS.BAND3_GAIN=value;        
            case 'BAND4_GAIN'    
                MTL.PRODUCT_PARAMETERS.BAND4_GAIN=value;        
            case 'BAND5_GAIN'    
                MTL.PRODUCT_PARAMETERS.BAND5_GAIN=value;        
            case 'BAND6_GAIN1'    
                MTL.PRODUCT_PARAMETERS.BAND6_GAIN1=value;        
            case 'BAND6_GAIN2'    
                MTL.PRODUCT_PARAMETERS.BAND6_GAIN2=value;        
            case 'BAND7_GAIN'    
                MTL.PRODUCT_PARAMETERS.BAND7_GAIN=value;        
            case 'BAND8_GAIN'
                MTL.PRODUCT_PARAMETERS.BAND8_GAIN=value;        
            case 'BAND1_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND1_GAIN_CHANGE=value;        
            case 'BAND2_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND2_GAIN_CHANGE=value;        
            case 'BAND3_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND3_GAIN_CHANGE=value;        
            case 'BAND4_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND4_GAIN_CHANGE=value;        
            case 'BAND5_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND5_GAIN_CHANGE=value;        
            case 'BAND6_GAIN_CHANGE1'
                MTL.PRODUCT_PARAMETERS.BAND6_GAIN_CHANGE1=value;        
            case 'BAND6_GAIN_CHANGE2'
                MTL.PRODUCT_PARAMETERS.BAND6_GAIN_CHANGE2=value;        
            case 'BAND7_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND7_GAIN_CHANGE=value;        
            case 'BAND8_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND8_GAIN_CHANGE=value;        
            case 'BAND1_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND1_SL_GAIN_CHANGE=value;        
            case 'BAND2_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND2_SL_GAIN_CHANGE=value;        
            case 'BAND3_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND3_SL_GAIN_CHANGE=value;        
            case 'BAND4_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND4_SL_GAIN_CHANGE=value;        
            case 'BAND5_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND5_SL_GAIN_CHANGE=value;        
            case 'BAND6_SL_GAIN_CHANGE1'
                MTL.PRODUCT_PARAMETERS.BAND6_SL_GAIN_CHANGE1=value;        
            case 'BAND6_SL_GAIN_CHANGE2'
                MTL.PRODUCT_PARAMETERS.BAND6_SL_GAIN_CHANGE2=value;        
            case 'BAND7_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND7_SL_GAIN_CHANGE=value;        
            case 'BAND8_SL_GAIN_CHANGE'
                MTL.PRODUCT_PARAMETERS.BAND8_SL_GAIN_CHANGE=value;        
            case 'SUN_AZIMUTH'
                MTL.PRODUCT_PARAMETERS.SUN_AZIMUTH=value;   
            case 'SUN_ELEVATION'
                MTL.PRODUCT_PARAMETERS.SUN_ELEVATION=value;   
            case 'OUTPUT_FORMAT'
                MTL.PRODUCT_PARAMETERS.OUTPUT_FORMAT=value;   

            %GROUP = CORRECTIONS_APPLIED
            case 'STRIPING_BAND1'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND1=value;
            case 'STRIPING_BAND2'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND2=value;
            case 'STRIPING_BAND3'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND3=value;
            case 'STRIPING_BAND4'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND4=value;
            case 'STRIPING_BAND5'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND5=value;
            case 'STRIPING_BAND61'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND61=value;
            case 'STRIPING_BAND62'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND62=value;
            case 'STRIPING_BAND7'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND7=value;
            case 'STRIPING_BAND8'
                MTL.CORRECTIONS_APPLIED.STRIPING_BAND8=value;
            case 'BANDING'
                MTL.CORRECTIONS_APPLIED.BANDING=value;
            case 'COHERENT_NOISE'
                MTL.CORRECTIONS_APPLIED.COHERENT_NOISE=value;
            case 'MEMORY_EFFECT'
                MTL.CORRECTIONS_APPLIED.MEMORY_EFFECT=value;
            case 'SCAN_CORRELATED_SHIFT'
                MTL.CORRECTIONS_APPLIED.SCAN_CORRELATED_SHIFT=value;
            case 'INOPERABLE_DETECTORS'
                MTL.CORRECTIONS_APPLIED.INOPERABLE_DETECTORS=value;
            case 'DROPPED_LINES'
                MTL.CORRECTIONS_APPLIED.DROPPED_LINES=value;


            %GROUP = PROJECTION_PARAMETERS
            case 'REFERENCE_DATUM'
                MTL.PROJECTION_PARAMETERS.REFERENCE_DATUM=value;
            case 'REFERENCE_ELLIPSOID'
                MTL.PROJECTION_PARAMETERS.REFERENCE_ELLIPSOID=value;
            case 'GRID_CELL_SIZE_PAN'
                MTL.PROJECTION_PARAMETERS.GRID_CELL_SIZE_PAN=value;
            case 'GRID_CELL_SIZE_THM'
                MTL.PROJECTION_PARAMETERS.GRID_CELL_SIZE_THM=value;
            case 'GRID_CELL_SIZE_REF'
                MTL.PROJECTION_PARAMETERS.GRID_CELL_SIZE_REF=value;
            case 'ORIENTATION'
                MTL.PROJECTION_PARAMETERS.ORIENTATION=value;
            case 'RESAMPLING_OPTION'
                MTL.PROJECTION_PARAMETERS.RESAMPLING_OPTION=value;
            case 'SCAN_GAP_INTERPOLATION'
                MTL.PROJECTION_PARAMETERS.SCAN_GAP_INTERPOLATION=value;
            case 'MAP_PROJECTION'
                MTL.PROJECTION_PARAMETERS.MAP_PROJECTION=value;
    
            %GROUP = UTM_PARAMETERS
            case 'ZONE_NUMBER'
                MTL.UTM_PARAMETERS.ZONE_NUMBER=value;
                
        end
        
        str_in=fgetl(fin);
    end
    
    MTL_list(cnt,1)=MTL;
    fclose(fin);
    
end